public class PartiallyUnrolledFindWhileBreak {

  /*@ model_behavior
    @ accessible \nothing;
    model static boolean lemma(int[] arr, int elem, int i) {
      return (\forall int k; (k >= 1 + i || k <= -1 || arr[k] != elem))
         &&  (\forall int k; (k == 1 + i || k == 2 + i || k == 3 + i || k == 4 + i ==> arr[k] != elem))
         ==> (\forall int k; (k >= 5 + i || k <= -1 || arr[k] != elem));
    }
    @*/

  /*@ public normal_behavior
    @ requires arr != null;
    @ ensures \result == -1 || arr[\result] == elem;
    @*/
  public static int partiallyUnrolledFindWhileBreak(int[] arr, int elem) {
    int i = -1, res = -1;
    /*@ loop_invariant
      @   (\forall int k; k <= i && k >= 0; arr[k] != elem) &&
      @   (res == -1 || arr[res] == elem) &&
      @   i >= -1 && i <= arr.length;
      @ decreases arr.length - i + 1;
      @ assignable i, res;
      @*/
    while (++i < arr.length) {
      // The following cut is needed to prove the goal
      // for the "i += 3; continue;" case:
      //
      /*
         \forall int k; (k >= 1 + i_0 | k <= -1 | !arr_0[k] = elem)
           & \forall int k; (k = 1 + i_0 | k = 2 + i_0 | k = 3 + i_0 | k = 4 + i_0 -> !arr_0[k] = elem)    
         ->
         \forall int k; (k >= 5 + i_0 | k <= -1 | !arr_0[k] = elem)
      */
      // Alternatively, there is a block contract for a
      // dummy block below; this contract introduces the
      // lemma. In the proof, one still needs to do two
      // easy quantifier instantiations.
      //
      if (i + 3 < arr.length) {
        if (arr[i] == elem) {
          res = i;
          break;
        } else if (arr[i + 1] == elem) {
          res = i + 1;
          break;
        } else if (arr[i + 2] == elem) {
          res = i + 2;
          break;
        } else if (arr[i + 3] == elem) {
          res = i + 3;
          break;
        } else {
          /*@ normal_behavior
            @ ensures lemma(arr, elem, i - 1);
            @ assignable \strictly_nothing;
            @*/
          {
            // this is just for KeY not to remove the
            // block as "empty block" and to ignore the
            // contract that we want to use as a lemma.
            // A "JML assert" would be better, but this
            // is currently not supported by KeY.
            int placeholder;
          }
          i += 3;
          continue;
        }
      }

      if (arr[i] == elem) {
        res = i;
        break;
      }
    }

    return res;
  }
}
